# -*- coding: utf-8 -*-

import subprocess
import threading


class Audio(object):
    def __init__(self, rate=16000, frames_size=160, channels=1, device_name='default'):
        self.rate = rate
        self.frames_size = frames_size
        self.channels = channels
        self.device_name = device_name
        self.done = False
        self.thread = None
        self.sinks = []

    def run(self):
        cmd = [
            'arecord',
            '-t', 'raw',
            '-f', 'S16_LE',
            '-c', str(self.channels),
            '-r', str(self.rate),
            '-D', self.device_name,
            '-q'
        ]
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE)

        frames_bytes = int(self.frames_size * self.channels * 2)
        while not self.done:
            audio = process.stdout.read(frames_bytes)
            for sink in self.sinks:
                sink.put(audio)

        process.kill()

    def start(self):
        self.done = False
        self.thread = threading.Thread(target=self.run)
        self.thread.daemon = True
        self.thread.start()

    def stop(self):
        self.done = True
        if self.thread and self.thread.is_alive():
            self.thread.join(timeout=3)

    def link(self, sink):
        if hasattr(sink, 'put') and callable(sink.put):
            self.sinks.append(sink)
        else:
            raise ValueError('Not implement put() method')

    def unlink(self, sink):
        self.sinks.remove(sink)



# -*- coding: utf-8 -*-

"""Player
support mpv, mpg123 and gstreamer 1.0
It likes gstreamer 1.0 > mpv > mpg123
We can specify a player using environment variable PLAYER (mpv, mpg123, gstreamer)
"""

import os

player_option = os.getenv('PLAYER', 'default').lower()

if player_option.find('mpv') >= 0:
    from .mpv_player import Player
elif player_option.find('mpg123') >= 0:
    from .mpg123_player import Player
elif player_option.find('gstreamer') >= 0:
    from .gstreamer_player import Player
else:
    try:
        from .gstreamer_player import Player
    except ImportError:
        if os.system('which mpv >/dev/null') == 0:
            from .mpv_player import Player
        elif os.system('which mpg123 >/dev/null') == 0:
            from .mpg123_player import Player
        else:
            raise ImportError('No player available, install one of the players: gstreamer, mpv and mpg123 first')

__all__ = ['Player']



# -*- coding: utf-8 -*-

import os

recorder_option = os.getenv('RECORDER', 'default').lower()

if recorder_option.find('pyaudio') >= 0 or os.system('which arecord >/dev/null') != 0:
    from .pyaudio_recorder import Audio
else:
    from .alsa_recorder import Audio

__all__ = ['Audio']



# -*- coding: utf-8 -*-

"""https://developer.amazon.com/public/solutions/alexa/alexa-voice-service/reference/audioplayer"""

import base64
import hashlib
import logging
import os
import tempfile
import uuid
from contextlib import closing

import requests

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

from avs.player import Player

logger = logging.getLogger('AudioPlayer')


# TODO: parse M3U8 and PLSv2
# refer to https://github.com/alexa/avs-device-sdk/blob/master/PlaylistParser/src/PlaylistParser.cpp
def get_audio_url(audio_url, timeout=3000):
    timeout = timeout / 1000.
    if audio_url.startswith('cid:'):
        filename = base64.urlsafe_b64encode(audio_url[4:])
        filename = hashlib.md5(filename).hexdigest()
        mp3_file = os.path.join(tempfile.gettempdir(), filename + '.mp3')
        if os.path.isfile(mp3_file):
            return 'file://{}'.format(mp3_file)
        else:
            logger.warn('Unable to parse {}'.format(audio_url))
            return None

    if audio_url.find('radiotime.com') >= 0:
        logger.debug('parse TuneIn audio stream: {}'.format(audio_url))

        try:
            response = requests.get(audio_url, timeout=timeout)
            lines = response.content.decode().split('\n')
            # audio/x-mpegurl; charset=utf-8
            content_type = response.headers['Content-Type']
            logger.debug(content_type)
            logger.debug(lines)
            if lines and lines[0]:
                audio_url = lines[0]
                logger.debug('Set audio_url to {}'.format(audio_url))
                if content_type.find('audio/x-mpegurl'):
                    return audio_url
        except Exception:
            pass

    extension = urlparse(audio_url).path[-4:]
    if extension in ['.mp3', '.wma']:
        logger.debug('Found audio stream {}'.format(audio_url))
        return audio_url

    logger.debug('Parse stream: {}'.format(audio_url))
    with closing(requests.get(audio_url, timeout=timeout, stream=True)) as response:
        content_type = response.headers['Content-Type'] or ''
        if content_type.find('pls') >= 0:
            try:
                logger.debug('parsing playlist: {}'.format(audio_url))
                lines = response.content.decode().split('\n')
                logger.debug(lines)
                for line in lines:
                    if line.find('File') == 0:
                        audio_url = lines[2].split('=', 2)[1]
                        logger.debug('Set audio_url to {}'.format(audio_url))
                        break
            except Exception:
                pass

    return audio_url


class AudioPlayer(object):
    STATES = {'IDLE', 'PLAYING', 'STOPPED', 'PAUSED', 'BUFFER_UNDERRUN', 'FINISHED'}

    def __init__(self, alexa):
        self.alexa = alexa
        self.token = ''
        self.state = 'IDLE'

        self.player = Player()
        self.player.add_callback('eos', self.PlaybackFinished)
        self.player.add_callback('error', self.PlaybackFailed)

    # {
    #     "directive": {
    #         "header": {
    #             "namespace": "AudioPlayer",
    #             "name": "Play",
    #             "messageId": "{{STRING}}",
    #             "dialogRequestId": "{{STRING}}"
    #         },
    #         "payload": {
    #             "playBehavior": "{{STRING}}",
    #             "audioItem": {
    #                 "audioItemId": "{{STRING}}",
    #                 "stream": {
    #                     "url": "{{STRING}}",
    #                     "streamFormat": "AUDIO_MPEG"
    #                     "offsetInMilliseconds": {{LONG}},
    #                     "expiryTime": "{{STRING}}",
    #                     "progressReport": {
    #                         "progressReportDelayInMilliseconds": {{LONG}},
    #                         "progressReportIntervalInMilliseconds": {{LONG}}
    #                     },
    #                     "token": "{{STRING}}",
    #                     "expectedPreviousToken": "{{STRING}}"
    #                 }
    #             }
    #         }
    #     }
    # }
    def Play(self, directive):
        if self.alexa.SpeechSynthesizer.state == 'PLAYING':
            self.alexa.SpeechSynthesizer.wait()

        behavior = directive['payload']['playBehavior']
        self.token = directive['payload']['audioItem']['stream']['token']
        audio_url = get_audio_url(directive['payload']['audioItem']['stream']['url'])

        self.player.play(audio_url)
        self.PlaybackStarted()

        logger.info('audio player is playing')

    def PlaybackStarted(self):
        self.state = 'PLAYING'

        event = {
            "header": {
                "namespace": "AudioPlayer",
                "name": "PlaybackStarted",
                "messageId": uuid.uuid4().hex
            },
            "payload": {
                "token": self.token,
                "offsetInMilliseconds": self.player.position
            }
        }

        self.alexa.send_event(event)

    def PlaybackNearlyFinished(self):
        event = {
            "header": {
                "namespace": "AudioPlayer",
                "name": "PlaybackNearlyFinished",
                "messageId": uuid.uuid4().hex
            },
            "payload": {
                "token": self.token,
                "offsetInMilliseconds": self.player.position
            }
        }
        self.alexa.send_event(event)

    def ProgressReportDelayElapsed(self):
        pass

    def ProgressReportIntervalElapsed(self):
        pass

    def PlaybackStutterStarted(self):
        pass

    def PlaybackStutterFinished(self):
        pass

    def PlaybackFinished(self):
        self.state = 'FINISHED'
        logger.info('playback is finished')

        event = {
            "header": {
                "namespace": "AudioPlayer",
                "name": "PlaybackFinished",
                "messageId": uuid.uuid4().hex
            },
            "payload": {
                "token": self.token,
                "offsetInMilliseconds": self.player.position
            }
        }
        self.alexa.send_event(event)

    def PlaybackFailed(self):
        self.state = 'STOPPED'

    # {
    #     "directive": {
    #         "header": {
    #             "namespace": "AudioPlayer",
    #             "name": "Stop",
    #             "messageId": "{{STRING}}",
    #             "dialogRequestId": "{{STRING}}"
    #         },
    #         "payload": {
    #         }
    #     }
    # }
    def Stop(self, directive):
        if self.state == 'PLAYING' or self.state == 'PAUSED':
            self.player.stop()
            self.PlaybackStopped()

            logger.info('audio player is stoped')

    def PlaybackStopped(self):
        self.state = 'STOPPED'
        event = {
            "header": {
                "namespace": "AudioPlayer",
                "name": "PlaybackStopped",
                "messageId": uuid.uuid4().hex
            },
            "payload": {
                "token": self.token,
                "offsetInMilliseconds": self.player.position
            }
        }
        self.alexa.send_event(event)

    def pause(self):
        self.player.pause()
        self.PlaybackPaused()

        logger.info('audio player is paused')

    def PlaybackPaused(self):
        self.state = 'PAUSED'
        event = {
            "header": {
                "namespace": "AudioPlayer",
                "name": "PlaybackPaused",
                "messageId": uuid.uuid4().hex
            },
            "payload": {
                "token": self.token,
                "offsetInMilliseconds": self.player.position
            }
        }
        self.alexa.send_event(event)

    def resume(self):
        self.player.resume()
        self.PlaybackResumed()

        logger.info('audio player is resumed')

    def PlaybackResumed(self):
        self.state = 'PLAYING'
        event = {
            "header": {
                "namespace": "AudioPlayer",
                "name": "PlaybackResumed",
                "messageId": uuid.uuid4().hex
            },
            "payload": {
                "token": self.token,
                "offsetInMilliseconds": self.player.position
            }
        }
        self.alexa.send_event(event)

    # {
    #     "directive": {
    #         "header": {
    #             "namespace": "AudioPlayer",
    #             "name": "ClearQueue",
    #             "messageId": "{{STRING}}",
    #             "dialogRequestId": "{{STRING}}"
    #         },
    #         "payload": {
    #             "clearBehavior": "{{STRING}}"
    #         }
    #     }
    # }
    def ClearQueue(self, directive):
        self.PlaybackQueueCleared()
        behavior = directive['payload']['clearBehavior']
        if behavior == 'CLEAR_ALL':
            self.player.stop()
        elif behavior == 'CLEAR_ENQUEUED':
            pass

    def PlaybackQueueCleared(self):
        event = {
            "header": {
                "namespace": "AudioPlayer",
                "name": "PlaybackQueueCleared",
                "messageId": uuid.uuid4().hex
            },
            "payload": {}
        }
        self.alexa.send_event(event)

    def StreamMetadataExtracted(self):
        pass

    @property
    def context(self):
        if self.state != 'PLAYING':
            offset = 0
        else:
            offset = self.player.position

        return {
            "header": {
                "namespace": "AudioPlayer",
                "name": "PlaybackState"
            },
            "payload": {
                "token": self.token,
                "offsetInMilliseconds": offset,
                "playerActivity": self.state
            }
        }



# -*- coding: utf-8 -*-

"""
Hands-free alexa with respeaker using pocketsphinx to search keyword

It depends on respeaker python library (https://github.com/respeaker/respeaker_python_library)
"""

import audioop
import os
import signal
import time

from avs.mic import Audio
from avs.player import Player


class RMS(object):
    def __init__(self):
        pass

    def put(self, data):
        print('RMS: {}'.format(audioop.rms(data, 2)))


def main():
    audio = Audio(frames_size=1600)
    rms = RMS()

    audio.link(rms)

    audio.start()

    alarm = os.path.realpath(os.path.join(os.path.dirname(__file__), 'resources/alarm.mp3'))
    alarm_uri = 'file://{}'.format(alarm)

    player1 = Player()
    player2 = Player()

    is_quit = []

    def signal_handler(signal, frame):
        print('Quit')
        is_quit.append(True)

    signal.signal(signal.SIGINT, signal_handler)

    while not is_quit:
        player1.play(alarm_uri)
        time.sleep(1)
        player1.pause()
        player2.play(alarm_uri)
        time.sleep(3)
        player2.pause()

    audio.stop()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""
Hands-free alexa with respeaker using pocketsphinx to search keyword

It depends on respeaker python library (https://github.com/respeaker/respeaker_python_library)
"""

import logging
import signal
import sys
import threading
import time

if sys.version_info < (3, 0):
    import Queue as queue
else:
    import queue

from avs.alexa import Alexa
from avs.mic import Audio
from respeaker.pixel_ring import pixel_ring

logger = logging.getLogger(__file__)
logging.basicConfig(level=logging.INFO)


class KWS(object):
    def __init__(self):
        self.queue = queue.Queue()

        self.sinks = []
        self._callback = None

        self.done = False

    def put(self, data):
        self.queue.put(data)

    def start(self):
        self.done = False
        thread = threading.Thread(target=self.run)
        thread.daemon = True
        thread.start()

    def stop(self):
        self.done = True

    def link(self, sink):
        if hasattr(sink, 'put') and callable(sink.put):
            self.sinks.append(sink)
        else:
            raise ValueError('Not implement put() method')

    def unlink(self, sink):
        self.sinks.remove(sink)

    def set_callback(self, callback):
        self._callback = callback

    def run(self):
        from respeaker.microphone import Microphone

        decoder = Microphone.create_decoder()
        decoder.start_utt()

        while not self.done:
            chunk = self.queue.get()
            decoder.process_raw(chunk, False, False)
            hypothesis = decoder.hyp()
            if hypothesis:
                keyword = hypothesis.hypstr
                logger.info('Detected {}'.format(keyword))

                if callable(self._callback):
                    self._callback(keyword)

                decoder.end_utt()
                decoder.start_utt()

            for sink in self.sinks:
                sink.put(chunk)


def main():
    config = None if len(sys.argv) < 2 else sys.argv[1]

    audio = Audio(frames_size=1600)
    kws = KWS()
    alexa = Alexa(config)

    def speak():
        pixel_ring.speak(10, 0)

    alexa.state_listener.on_listening = pixel_ring.listen
    alexa.state_listener.on_thinking = pixel_ring.wait
    alexa.state_listener.on_speaking = speak
    alexa.state_listener.on_finished = pixel_ring.off

    audio.link(kws)
    kws.link(alexa)

    def wakeup(keyword):
        if keyword.find('alexa') >= 0:
            alexa.listen()

    kws.set_callback(wakeup)

    alexa.start()
    kws.start()
    audio.start()

    is_quit = threading.Event()

    def signal_handler(signal, frame):
        print('Quit')
        is_quit.set()

    signal.signal(signal.SIGINT, signal_handler)

    while not is_quit.is_set():
        time.sleep(1)

    alexa.stop()
    kws.stop()
    audio.stop()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""Player using MPV"""

import os
import subprocess
import threading

if os.system('which mpv >/dev/null') != 0:
    raise ImportError('mpv not found, install it first')


class Player(object):
    def __init__(self):
        self.callbacks = {}
        self.process = None
        self.state = 'NULL'
        self.audio = None
        self.tty = None

        self.event = threading.Event()
        t = threading.Thread(target=self._run)
        t.daemon = True
        t.start()

    def _run(self):
        while True:
            self.event.wait()
            self.event.clear()
            print('Playing {}'.format(self.audio))

            master, slave = os.openpty()
            self.process = subprocess.Popen(['mpv', '--no-video', self.audio], stdin=master)
            self.tty = slave

            self.process.wait()
            print('Finished {}'.format(self.audio))

            if not self.event.is_set():
                self.on_eos()

    def play(self, uri):
        self.audio = uri
        self.event.set()

        if self.process and self.process.poll() == None:
            os.write(self.tty, b'q')

        self.state = 'PLAYING'

    def stop(self):
        if self.process and self.process.poll() == None:
            os.write(self.tty, b'q')
        self.state = 'NULL'

    def pause(self):
        if self.state == 'PLAYING':
            self.state = 'PAUSED'
            os.write(self.tty, ' ')

        print('pause()')

    def resume(self):
        if self.state == 'PAUSED':
            self.state = 'PLAYING'
            os.write(self.tty, b' ')

    # name: {eos, ...}
    def add_callback(self, name, callback):
        if not callable(callback):
            return

        self.callbacks[name] = callback

    def on_eos(self):
        self.state = 'NULL'
        if 'eos' in self.callbacks:
            self.callbacks['eos']()

    @property
    def duration(self):
        return 0

    @property
    def position(self):
        return 0



import datetime
import json
import os
import time
import uuid

import click
import requests
import tornado.httpserver
import tornado.ioloop
import tornado.web

import avs.config


class MainHandler(tornado.web.RequestHandler):
    def initialize(self, config, output):
        self.config = config
        self.output = output

    @tornado.web.asynchronous
    def get(self):
        redirect_uri = self.request.protocol + "://" + self.request.host + "/authresponse"
        if self.request.path == '/authresponse':
            code = self.get_argument("code")
            payload = {
                "client_id": self.config['client_id'],
                "client_secret": self.config['client_secret'],
                "code": code,
                "grant_type": "authorization_code",
                "redirect_uri": redirect_uri
            }

            if self.config['host_url'] == 'dueros-h2.baidu.com':
                token_url = 'https://openapi.baidu.com/oauth/2.0/token'
                message = 'Succeed to login Baidu DuerOS'
            else:
                token_url = 'https://api.amazon.com/auth/o2/token'
                message = 'Succeed to login Amazon Alexa Voice Service'

            r = requests.post(token_url, data=payload)
            config = r.json()
            self.config['refresh_token'] = config['refresh_token']

            if 'access_token' in config:
                date_format = "%a %b %d %H:%M:%S %Y"
                expiry_time = datetime.datetime.utcnow() + datetime.timedelta(seconds=config['expires_in'])
                self.config['expiry'] = expiry_time.strftime(date_format)
                self.config['access_token'] = config['access_token']

            print(json.dumps(self.config, indent=4))
            avs.config.save(self.config, configfile=self.output)

            self.write(message)
            self.finish()
            tornado.ioloop.IOLoop.instance().stop()
        elif self.request.path == '/alexa':
            self.alexa_oauth()
        elif self.request.path == '/dueros':
            self.dueros_oauth()
        elif self.request.path == '/':
            index_html = os.path.realpath(os.path.join(os.path.dirname(__file__), 'resources/web/index.html'))
            with open(index_html) as f:
                self.write(f.read())
                self.finish()

    def alexa_oauth(self):
        if 'client_secret' not in self.config:
            self.config.update(avs.config.alexa())
        if 'dueros-device-id' in self.config:
            del self.config['dueros-device-id']
            self.config.update(avs.config.alexa())

        oauth_url = 'https://www.amazon.com/ap/oa'
        redirect_uri = self.request.protocol + "://" + self.request.host + "/authresponse"

        scope_data = json.dumps({
            "alexa:all": {
                "productID": self.config['product_id'],
                "productInstanceAttributes": {
                    "deviceSerialNumber": uuid.uuid4().hex
                }
            }
        })
        payload = {
            "client_id": self.config['client_id'],
            "scope": "alexa:all",
            "scope_data": scope_data,
            "response_type": "code",
            "redirect_uri": redirect_uri
        }

        req = requests.Request('GET', oauth_url, params=payload)
        p = req.prepare()
        self.redirect(p.url)

    def dueros_oauth(self):
        if 'client_secret' not in self.config:
            self.config.update(avs.config.dueros())
        
        oauth_url = 'https://openapi.baidu.com/oauth/2.0/authorize'
        redirect_uri = self.request.protocol + "://" + self.request.host + "/authresponse"

        payload = {
            "client_id": self.config["client_id"],
            "scope": "basic",
            "response_type": "code",
            "redirect_uri": redirect_uri
        }

        req = requests.Request('GET', oauth_url, params=payload)
        p = req.prepare()
        self.redirect(p.url)


def open_webbrowser():
    try:
        import webbrowser
    except ImportError:
        print('Go to http://{your device IP}:3000 to start')
        return

    time.sleep(0.1)
    print("A web page should is opened. If not, go to http://127.0.0.1:3000 to start")
    webbrowser.open('http://127.0.0.1:3000')


def auth(config, output):
    import threading
    threading.Thread(target=open_webbrowser).start()

    config = avs.config.load(config) if config else {}

    application = tornado.web.Application([(r".*", MainHandler, dict(config=config, output=output))])
    http_server = tornado.httpserver.HTTPServer(application)
    http_server.listen(3000)
    tornado.ioloop.IOLoop.instance().start()
    tornado.ioloop.IOLoop.instance().close()

@click.command()
@click.option('--config', '-c', help='configuration json file with product_id, client_id and client_secret')
@click.option('--output', '-o', default=avs.config.DEFAULT_CONFIG_FILE, help='output json file with refresh token')
def main(config, output):
    auth(config, output)


if __name__ == '__main__':
    main()



class Speaker(object):
    def __init__(self, alexa):
        self.alexa = alexa
        self.volume = 50
        self.muted = False

        self.set_volume_cb = None
        self.get_volume_cb = None
        self.set_mute_cb = None

    def SetVolume(self, directive):
        vol = directive['payload']['volume']
        if self.set_volume_cb:
            self.set_volume_cb(vol)
            self.volume = vol

    def CallbackSetVolume(self, func):
        self.set_volume_cb = func

    def AdjustVolume(self, directive):
        vol = directive['payload']['volume']
        if self.get_volume_cb:
            self.volume = self.get_volume_cb()
        
        self.volume += vol
        if (self.volume > 100):
            self.volume = 100
        elif (self.volume < 0):
            self.volume = 0

        if self.set_volume_cb:
            self.set_volume_cb(self.volume)
            
    
    def CallbackGetVolume(self, func):
        self.get_volume_cb = func

    def VolumeChanged(self):
        event = {
            "event": {
                "header": {
                    "namespace": "Speaker",
                    "name": "VolumeChanged",
                    "messageId": "{{STRING}}"
                },
                "payload": {
                    "volume": self.volume,
                    "muted": self.muted
                }
            }
        }
        self.alexa.send_event(event)

    def SetMute(self, directive):
        muted = directive["payload"]["mute"]
        if self.set_mute_cb:
            self.set_mute_cb(muted)
            self.muted = muted

    
    def CallbackSetMute(self, func):
        self.set_mute_cb = func

    def MuteChanged(self):
        event = {
            "event": {
                "header": {
                    "namespace": "Speaker",
                    "name": "MuteChanged",
                    "messageId": "{{STRING}}"
                },
                "payload": {
                    "volume": self.volume,
                    "muted": self.muted
                }
            }
        }
        self.alexa.send_event(event)

    @property
    def context(self):
        return {
            "header": {
                "namespace": "Speaker",
                "name": "VolumeState"
            },
            "payload": {
                "volume": self.volume,
                "muted": self.muted,
            }
        }




# -*- coding: utf-8 -*-

"""Player using gstreamer."""

import gi
import threading
import time

gi.require_version('Gst', '1.0')
from gi.repository import Gst, GLib, GObject


def setup_message_handler():
    GObject.threads_init()
    Gst.init(None)
    loop = GLib.MainLoop()

    t = threading.Thread(target=loop.run)
    t.daemon = True
    t.start()

    return t


class Player(object):
    message_handler = None

    def __init__(self):
        if Player.message_handler is None:
            Player.message_handler = setup_message_handler()

        self.callbacks = {}
        self.player = Gst.ElementFactory.make("playbin", "player")
        bus = self.player.get_bus()
        bus.add_signal_watch()
        bus.connect('message', self.on_message)

        # bus.enable_sync_message_emission()
        # bus.connect('sync-message::eos', self.on_eos)

    def play(self, uri):
        self.player.set_state(Gst.State.NULL)
        self.player.set_property('uri', uri)
        self.player.set_state(Gst.State.PLAYING)

    def stop(self):
        self.player.set_state(Gst.State.NULL)

    def pause(self):
        self.player.set_state(Gst.State.PAUSED)

    def resume(self):
        self.player.set_state(Gst.State.PLAYING)

    # name: {eos, error, ...}
    def add_callback(self, name, callback):
        if not callable(callback):
            return

        self.callbacks[name] = callback

    def on_message(self, bus, message):
        if message.type == Gst.MessageType.EOS:
            self.player.set_state(Gst.State.NULL)
            if 'eos' in self.callbacks:
                self.callbacks['eos']()
        elif message.type == Gst.MessageType.ERROR:
            self.player.set_state(Gst.State.NULL)
            if 'error' in self.callbacks:
                self.callbacks['error']()
                # else:
                #     print(message.type)

    @property
    def duration(self):
        for _ in range(10):
            success, duration = self.player.query_duration(Gst.Format.TIME)
            if success:
                break
            time.sleep(0.1)

        return int(duration / Gst.MSECOND)

    @property
    def position(self):
        for _ in range(10):
            success, position = self.player.query_position(Gst.Format.TIME)
            if success:
                break
            time.sleep(0.1)

        return int(position / Gst.MSECOND)

    @property
    def state(self):
        # GST_STATE_VOID_PENDING        no pending state.
        # GST_STATE_NULL                the NULL state or initial state of an element.
        # GST_STATE_READY               the element is ready to go to PAUSED.
        # GST_STATE_PAUSED              the element is PAUSED, it is ready to accept and process data.
        #                               Sink elements however only accept one buffer and then block.
        # GST_STATE_PLAYING             the element is PLAYING, the GstClock is running and the data is flowing.
        _, state, _ = self.player.get_state(Gst.SECOND)
        return 'FINISHED' if state != Gst.State.PLAYING else 'PLAYING'



# -*- coding: utf-8 -*-

import json
import os
import sys
import uuid

DEFAULT_CONFIG_FILE = os.path.join(os.path.expanduser('~'), '.avs.json')


def load(configfile=DEFAULT_CONFIG_FILE):
    if configfile is None:
        configfile = DEFAULT_CONFIG_FILE
        
    if not os.path.isfile(configfile):
        raise RuntimeError('No configuration file')

    with open(configfile, 'r') as f:
        config = json.load(f)
        require_keys = ['product_id', 'client_id', 'client_secret']
        for key in require_keys:
            if not ((key in config) and config[key]):
                raise RuntimeError('{} should include "{}"'.format(configfile, key))

    if ('host_url' not in config) or (not config['host_url']):
        config['host_url'] = 'avs-alexa-na.amazon.com'

    if config['host_url'] == 'dueros-h2.baidu.com':
        config['api'] = 'dcs/avs-compatible-v20160207'
        config['refresh_url'] = 'https://openapi.baidu.com/oauth/2.0/token'
    else:
        config['api'] = 'v20160207'
        config['refresh_url'] = 'https://api.amazon.com/auth/o2/token'

    return config


def save(config, configfile=None):
    if configfile is None:
        configfile = DEFAULT_CONFIG_FILE

    with open(configfile, 'w') as f:
        json.dump(config, f, indent=4)


def dueros():
    product_id = "xiaojing-" + uuid.uuid4().hex
    return {
        "dueros-device-id": product_id,
        "client_id": "lud6jnwdVFim4K4Zkoas3BkRVLvCO57Z",
        "host_url": "dueros-h2.baidu.com",
        "client_secret": "A017kke1GSSz7hp8Fj6ySoIWrnFraxf5",
        "product_id": product_id
    }

def alexa():
    return {
        "product_id": "ReSpeaker",
        "client_id": "amzn1.application-oa2-client.91b0cebd9074412cba1570a5dd03fc6e",
        "host_url": "avs-alexa-na.amazon.com",
        "client_secret": "fbd7a0e72953c1dd9a920670cf7f4115f694cd47c32a1513dc12a804c7f804e2"
    }


# -*- coding: utf-8 -*-


import base64
import cgi
import json
import logging
import os
import signal
import sys
import tempfile
import threading
import uuid

if sys.version_info < (3, 0):
    import Queue as queue
else:
    import queue

import requests
import datetime
import hyper

from avs.mic import Audio

from avs.interface.alerts import Alerts
from avs.interface.audio_player import AudioPlayer
from avs.interface.speaker import Speaker
from avs.interface.speech_recognizer import SpeechRecognizer
from avs.interface.speech_synthesizer import SpeechSynthesizer
from avs.interface.system import System
import avs.config
import avs.auth

logger = logging.getLogger(__name__)


class AlexaStateListener(object):
    def __init__(self):
        pass

    def on_ready(self):
        logger.info('on_ready')

    def on_disconnected(self):
        logger.info('on_disconnected')

    def on_listening(self):
        logger.info('on_listening')

    def on_thinking(self):
        logger.info('on_thinking')

    def on_speaking(self):
        logger.info('on_speaking')

    def on_finished(self):
        logger.info('on_finished')


class Alexa(object):
    API_VERSION = 'v20160207'

    def __init__(self, config=None):
        self.event_queue = queue.Queue()
        self.SpeechRecognizer = SpeechRecognizer(self)
        self.SpeechSynthesizer = SpeechSynthesizer(self)
        self.AudioPlayer = AudioPlayer(self)
        self.Speaker = Speaker(self)
        self.Alerts = Alerts(self)
        self.System = System(self)

        self.state_listener = AlexaStateListener()

        # put() will send audio to speech recognizer
        self.put = self.SpeechRecognizer.put

        # listen() will trigger SpeechRecognizer's Recognize event
        self.listen = self.SpeechRecognizer.Recognize

        self.done = False

        self.requests = requests.Session()

        self._configfile = config
        self._config = avs.config.load(configfile=config)

        self.last_activity = datetime.datetime.utcnow()
        self._ping_time = None

    def set_state_listener(self, listner):
        self.state_listener = listner

    def start(self):
        self.done = False

        t = threading.Thread(target=self.run)
        t.daemon = True
        t.start()

    def stop(self):
        self.done = True

    def send_event(self, event, listener=None, attachment=None):
        self.event_queue.put((event, listener, attachment))

    def run(self):
        while not self.done:
            try:
                self._run()
            except AttributeError as e:
                logger.exception(e)
                continue
            except hyper.http20.exceptions.StreamResetError as e:
                logger.exception(e)
                continue
            except ValueError as e:
                logging.exception(e)
                # failed to get an access token, exit
                sys.exit(1)
            except Exception as e:
                logging.exception(e)
                continue

    def _run(self):
        conn = hyper.HTTP20Connection('{}:443'.format(
            self._config['host_url']), force_proto='h2')

        headers = {'authorization': 'Bearer {}'.format(self.token)}
        if 'dueros-device-id' in self._config:
            headers['dueros-device-id'] = self._config['dueros-device-id']

        downchannel_id = conn.request(
            'GET', '/{}/directives'.format(self._config['api']), headers=headers)
        downchannel_response = conn.get_response(downchannel_id)
        if downchannel_response.status != 200:
            raise ValueError(
                "/directive requests returned {}".format(downchannel_response.status))

        _, pdict = cgi.parse_header(
            downchannel_response.headers['content-type'][0].decode('utf-8'))
        downchannel_boundary = '--{}'.format(pdict['boundary']).encode('utf-8')
        downchannel = conn.streams[downchannel_id]
        downchannel_buffer = b''
        eventchannel_boundary = 'seeed-voice-engine'

        # ping every 5 minutes (60 seconds early for latency) to maintain the connection
        self._ping_time = datetime.datetime.utcnow() + datetime.timedelta(seconds=240)
        self.event_queue.queue.clear()
        self.System.SynchronizeState()
        while not self.done:
            try:
                event, listener, attachment = self.event_queue.get(
                    timeout=0.25)
            except queue.Empty:
                event = None

            # we want to avoid blocking if the data wasn't for stream downchannel
            while conn._sock.can_read:
                conn._single_read()

            while downchannel.data:
                framebytes = downchannel._read_one_frame()
                downchannel_buffer = self._parse_response(
                    framebytes, downchannel_boundary, downchannel_buffer
                )

            if event is None:
                self._ping(conn)
                self.System.UserInactivityReport()
                continue

            headers = {
                ':method': 'POST',
                ':scheme': 'https',
                ':path': '/{}/events'.format(self._config['api']),
                'authorization': 'Bearer {}'.format(self.token),
                'content-type': 'multipart/form-data; boundary={}'.format(eventchannel_boundary)
            }
            if 'dueros-device-id' in self._config:
                headers['dueros-device-id'] = self._config['dueros-device-id']

            stream_id = conn.putrequest(headers[':method'], headers[':path'])
            default_headers = (':method', ':scheme', ':authority', ':path')
            for name, value in headers.items():
                is_default = name in default_headers
                conn.putheader(name, value, stream_id, replace=is_default)
            conn.endheaders(final=False, stream_id=stream_id)

            metadata = {
                'context': self.context,
                'event': event
            }
            # logger.info('metadata: {}'.format(json.dumps(metadata, indent=4)))

            json_part = '--{}\r\n'.format(eventchannel_boundary)
            json_part += 'Content-Disposition: form-data; name="metadata"\r\n'
            json_part += 'Content-Type: application/json; charset=UTF-8\r\n\r\n'
            json_part += json.dumps(metadata)

            conn.send(json_part.encode('utf-8'),
                      final=False, stream_id=stream_id)

            if attachment:
                attachment_header = '\r\n--{}\r\n'.format(
                    eventchannel_boundary)
                attachment_header += 'Content-Disposition: form-data; name="audio"\r\n'
                attachment_header += 'Content-Type: application/octet-stream\r\n\r\n'
                conn.send(attachment_header.encode('utf-8'),
                          final=False, stream_id=stream_id)

                # AVS_AUDIO_CHUNK_PREFERENCE = 320
                for chunk in attachment:
                    conn.send(chunk, final=False, stream_id=stream_id)

                    # check if StopCapture directive is received
                    while conn._sock.can_read:
                        conn._single_read()

                    while downchannel.data:
                        framebytes = downchannel._read_one_frame()
                        downchannel_buffer = self._parse_response(
                            framebytes, downchannel_boundary, downchannel_buffer
                        )

                self.last_activity = datetime.datetime.utcnow()

            end_part = '\r\n--{}--'.format(eventchannel_boundary)
            conn.send(end_part.encode('utf-8'),
                      final=True, stream_id=stream_id)

            logger.info("wait for response")
            response = conn.get_response(stream_id)
            logger.info("status code: %s", response.status)

            if response.status == 200:
                _, pdict = cgi.parse_header(
                    response.headers['content-type'][0].decode('utf-8'))
                boundary = '--{}'.format(pdict['boundary']).encode('utf-8')
                self._parse_response(response.read(), boundary)
            elif response.status == 204:
                pass
            else:
                logger.warning(response.headers)
                logger.warning(response.read())

            if listener and callable(listener):
                listener()

    def _parse_response(self, response, boundary, buffer=b''):
        directives = []
        blen = len(boundary)
        response = buffer + response
        while response:
            pos = response.find(boundary)
            if pos < 0:
                break

            # skip small data block
            if pos > blen:
                # a blank line is between parts
                parts = response[:pos - 2].split(b'\r\n\r\n', 1)
                if parts[0].find(b'application/json') >= 0:
                    metadata = json.loads(parts[1].decode('utf-8'))
                    if 'directive' in metadata:
                        directives.append(metadata['directive'])
                elif parts[0].find(b'application/octet-stream') >= 0:
                    for line in parts[0].splitlines():
                        name, value = line.split(b':', 1)
                        if name.lower() == b'content-id':
                            content_id = value.strip()[1:-1]
                            filename = base64.urlsafe_b64encode(content_id)[:8].decode('utf-8')
                            with open(os.path.join(tempfile.gettempdir(), '{}.mp3'.format(filename)), 'wb') as f:
                                f.write(parts[1])
                            logger.info('write audio to {}.mp3'.format(filename))
                            break

            response = response[pos + blen + 2:]

        for directive in directives:
            self._handle_directive(directive)

        return response

    def _handle_directive(self, directive):
        logger.info(json.dumps(directive, indent=4))
        try:
            namespace = directive['header']['namespace']
            name = directive['header']['name']
            if hasattr(self, namespace):
                interface = getattr(self, namespace)
                directive_func = getattr(interface, name, None)
                if directive_func:
                    directive_func(directive)
                else:
                    logger.info(
                        '{}.{} is not implemented yet'.format(namespace, name))
            else:
                logger.info('{} is not implemented yet'.format(namespace))

        except KeyError as e:
            logger.exception(e)
        except Exception as e:
            logger.exception(e)

    def _ping(self, connection):
        if datetime.datetime.utcnow() >= self._ping_time:
            connection.ping(uuid.uuid4().hex[:8])

            logger.debug('ping at {}'.format(
                datetime.datetime.utcnow().strftime("%a %b %d %H:%M:%S %Y")))

            # ping every 5 minutes (60 seconds early for latency) to maintain the connection
            self._ping_time = datetime.datetime.utcnow() + datetime.timedelta(seconds=240)

    @property
    def context(self):
        # return [self.SpeechRecognizer.context, self.SpeechSynthesizer.context,
        #                    self.AudioPlayer.context, self.Speaker.context, self.Alerts.context]
        return [self.SpeechSynthesizer.context, self.Speaker.context, self.AudioPlayer.context, self.Alerts.context]

    @property
    def token(self):
        date_format = "%a %b %d %H:%M:%S %Y"

        if 'access_token' in self._config:
            if 'expiry' in self._config:
                expiry = datetime.datetime.strptime(
                    self._config['expiry'], date_format)
                # refresh 60 seconds early to avoid chance of using expired access_token
                if (datetime.datetime.utcnow() - expiry) > datetime.timedelta(seconds=60):
                    logger.info("Refreshing access_token")
                else:
                    return self._config['access_token']

        payload = {
            'client_id': self._config['client_id'],
            'client_secret': self._config['client_secret'],
            'grant_type': 'refresh_token',
            'refresh_token': self._config['refresh_token']
        }

        response = None

        # try to request an access token 3 times
        for _ in range(3):
            try:
                response = self.requests.post(
                    self._config['refresh_url'], data=payload)
                if response.status_code != 200:
                    logger.warning(response.text)
                else:
                    break
            except Exception as e:
                logger.exception(e)
                continue

        if (response is None) or (not hasattr(response, 'status_code')) or response.status_code != 200:
            raise ValueError(
                "refresh token request returned {}".format(response.status))

        config = response.json()
        self._config['access_token'] = config['access_token']

        expiry_time = datetime.datetime.utcnow(
        ) + datetime.timedelta(seconds=config['expires_in'])
        self._config['expiry'] = expiry_time.strftime(date_format)
        logger.debug(json.dumps(self._config, indent=4))

        avs.config.save(self._config, configfile=self._configfile)

        return self._config['access_token']

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()


def main():
    logging.basicConfig(level=logging.INFO)

    config = avs.config.DEFAULT_CONFIG_FILE if len(sys.argv) < 2 else sys.argv[1]

    if not os.path.isfile(config):
        print('Login amazon alexa or baidu dueros first')
        avs.auth.auth(None, config)

    audio = Audio()
    alexa = Alexa(config)

    audio.link(alexa)

    alexa.start()
    audio.start()

    is_quit = threading.Event()

    def signal_handler(sig, frame):
        print('Quit')
        is_quit.set()

    signal.signal(signal.SIGINT, signal_handler)

    while True:
        try:
            input('press ENTER to talk\n')
        except SyntaxError:
            pass
        except NameError:
            pass

        if is_quit.is_set():
            break
            
        alexa.listen()

    alexa.stop()
    audio.stop()


if __name__ == '__main__':
    main()
